// doublehash finds map update operations whose key is the same as the
// that of a dominating map lookup operation.  These are opportunities
// for computing the key's hash only once.
//
// It does not find keys that are constants, or addresses of global
// variables.  (These could be hashed at compile- or link-time anwyay.)
//
// For example:
//
//	v, ok := m[k]
//	if !ok {
//		v = ...
//		m[k] = v
//	}
//
// Or:
//
//	v := m[k]
//	if v == nil {
//		v = ...
//		m[k] = v
//	}
//
// But not:
//
//	v := o.m[k]
//	if v == nil {
//		o.m[k] = v
//	}
//
// because the SSA's escape analysis is too weak to prove that the two
// references o.m are semantically identical.  Use -unsound to use
// looser matching for map variables, i.e., assume no intervening stores
// to o or o.m.  (There's no equivalent flag for keys of the form o.k.)
//
// See golang issue 5147 for context.
//
package main

import (
	"flag"
	"fmt"
	"go/build"
	"go/token"
	"log"
	"os"
	"runtime"

	"golang.org/x/tools/go/buildutil"
	"golang.org/x/tools/go/loader"
	"golang.org/x/tools/go/ssa"
	"golang.org/x/tools/go/ssa/ssautil"
	"golang.org/x/tools/go/types"
)

var unsound = flag.Bool("unsound", false,
	"Assume no intervening stores to map variables between lookup and update")

func main() {
	runtime.GOMAXPROCS(8)

	flag.Parse()

	conf := loader.Config{
		SourceImports: true,
		Build:         &build.Default,
		AllowErrors:   true,
	}

	pkgs := flag.Args()
	if len(pkgs) == 0 {
		fmt.Fprintln(os.Stderr, "Checking all packages in $GOPATH/$GOROOT...")
		pkgs = buildutil.AllPackages(conf.Build)
	}

	if _, err := conf.FromArgs(pkgs, true); err != nil {
		log.Fatalf("Failed to parse arguments: %v", err)
	}

	iprog, err := conf.Load()
	if err != nil {
		log.Fatalf("Load failed: %v", err)
	}

	prog := ssa.Create(iprog, 0)
	prog.BuildAll()

	initial := make(map[*types.Package]bool)
	for _, info := range iprog.InitialPackages() {
		initial[info.Pkg] = true
	}

	for fn := range ssautil.AllFunctions(prog) {
		// TODO(adonovan): make this predicate simpler.
		if fn.Pkg != nil && initial[fn.Pkg.Object] {
			for _, b := range fn.Blocks {
				for _, instr := range b.Instrs {
					if instr, ok := instr.(*ssa.Lookup); ok {
						checkLookup(instr)
					}
				}
			}
		}
	}
}

func checkLookup(lookup *ssa.Lookup) {
	if _, ok := lookup.X.Type().Underlying().(*types.Map); !ok {
		return // not a map lookup
	}
	if lookup.Index.Parent() != lookup.Parent() {
		// Index defined in an enclosing function,
		// or is a Const or Global.
		return
	}

	var updates []*ssa.MapUpdate
	for _, ref := range *lookup.Index.Referrers() {
		if update, ok := ref.(*ssa.MapUpdate); ok &&
			lookup.Index == update.Key &&
			same(lookup.X, update.Map) &&
			lookup.Block().Dominates(update.Block()) {
			updates = append(updates, update)
		}
	}
	if updates == nil {
		return // no dominated updates
	}

	fset := lookup.Parent().Prog.Fset
	fmt.Printf("%s:\tthis %s lookup\n",
		fset.Position(lookup.Pos()),
		types.TypeString(lookup.Parent().Pkg.Object, lookup.X.Type()))
	for _, update := range updates {
		fmt.Printf("%s:\t\tdominates this update with same map and key\n",
			fset.Position(update.Pos()))
	}
	fmt.Println()
}

func same(x, y ssa.Value) bool {
	if *unsound {
		if x, ok := x.(*ssa.UnOp); ok && x.Op == token.MUL {
			if y, ok := y.(*ssa.UnOp); ok && y.Op == token.MUL {
				return same(x.X, y.X)
			}
		}
		if x, ok := x.(*ssa.FieldAddr); ok {
			if y, ok := y.(*ssa.FieldAddr); ok {
				return x.Field == y.Field && same(x.X, y.X)
			}
		}
		if x, ok := x.(*ssa.IndexAddr); ok {
			if y, ok := y.(*ssa.IndexAddr); ok {
				return same(x.Index, y.Index) && same(x.X, y.X)
			}
		}
	}
	return x == y
}
